from typing import Collection

MESSAGE_MODULO_CHUNKS = 100
MIN_NUMBER_OF_MESSAGES = 100

PLACEHOLDER1 = 1  # An integer
PLACEHOLDER2 = []  # An iterable


def func1(chunk1) -> Collection:
    """
    Dummy function that represents a process. `chunks` is a function of `chunk1`
    """
    result = []
    return result


def getNumChunks(result) -> int:
    """
    Dummy function that represents a counting process.
    """
    numChunks = 1
    return numChunks


def getChunks(result) -> Collection:
    """
    A dummy function that represents a chunking process.
    """
    chunks = []
    return chunks


numChunks1 = PLACEHOLDER1
# >>> Calculate logging requency for loop layer 1
if numChunks1 < MESSAGE_MODULO_CHUNKS:
    moduloChunks1 = numChunks1
else:
    moduloChunks1 = round(numChunks1 / MESSAGE_MODULO_CHUNKS)
if numChunks1 / moduloChunks1 < MIN_NUMBER_OF_MESSAGES:
    moduloChunks1 = 1
else:
    pass
# <<< <<< <<< <<<
chunks = PLACEHOLDER2
for it1, chunk1 in enumerate(chunks, start=1):
    # >>> Set chunk verbosity for loop layer 1
    if it1 % moduloChunks1 == 0:
        allowLogging = True
    else:
        allowLogging = False
    # <<< <<< <<< <<<
    result1 = func1(chunk1)
    numChunks2 = getNumChunks(result1)
    chunks2 = getChunks(result1)
    # >>> Calculate logging requency for loop layer 2
    if numChunks2 < MESSAGE_MODULO_CHUNKS:
        moduloChunks2 = numChunks2
    else:
        moduloChunks2 = round(numChunks2 / MESSAGE_MODULO_CHUNKS)
    if numChunks2 / moduloChunks2 < MIN_NUMBER_OF_MESSAGES:
        moduloChunks2 = 1
    else:
        pass
    # <<< <<< <<< <<<
    for it2, chunk2 in enumerate(chunks2, start=1):
        # >>> Set chunk verbosity for loop layer 1
        if it2 % moduloChunks2 == 0:
            allowLogging = True
        else:
            allowLogging = False
        # <<< <<< <<< <<<



"""
"""

from pathlib import Path
from typing_extensions import Union


def preview_file(file_path: Path,
                 num_characters: int = 100):
    """
    """
    with open(file_path, "rb") as file:
        text = file.read(num_characters)
    print(text)


def replace_text(file_path_in: Path,
                 old_text: Union[bytes, str],
                 new_text: Union[bytes, str],
                 file_path_out: Path = None):
    """
    """
    # Read
    with open(file_path_in, "rb") as file:
        file_text = file.read()

    # Replace
    file_text_new_0 = file_text.replace(old_text, new_text)

    # Decode
    if isinstance(file_text, bytes):
        file_text_new_1 = file_text_new_0.decode()
    else:
        file_text_new_1 = file_text_new_0

    # Choose file path to write to.
    if file_path_out:
        file_path_out_ = file_path_out
    else:
        file_path_out_ = file_path_in

    # Write
    with open(file_path_out_, "w") as file:
        file.write(file_text_new_1)


def replace_text_wrapper(file_path_in: Path,
                         output_dir: Path,
                         old_text: Union[bytes, str],
                         new_text: Union[bytes, str]):
    """
    Wrapper for parallel processing.
    """
    file_path_in_name = file_path_in.name
    file_path_out = output_dir.joinpath(file_path_in_name)
    replace_text(file_path_in=file_path_in,
                 old_text=old_text,
                 new_text=new_text,
                 file_path_out=file_path_out)



"""
Split a column into multiple columns from a pre-existing table.
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
# Third-party packages
import pandas as pd
import pathos.pools as pp
# First-party packages
from drapi import __version__ as drapiVersion
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    makeDirPath)
from drapi.code.drapi.expandColumn import expandColumnWrapper

# Functions


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--PATHS",
                        nargs="+",
                        required=True,
                        help="The path to the file where the table is located. Usually a CSV or TSV file.")
    parser.add_argument("--COLUMN_TO_SPLIT",
                        required=True,
                        help="The label of the column to split.")
    parser.add_argument("--NAME_OF_NEW_COLUMNS",
                        nargs="+",
                        required=True,
                        help="A list of column names to give to the new columns.")
    parser.add_argument("--LOCATION_OF_NEW_COLUMNS",
                        nargs="+",
                        type=int,
                        required=True,
                        help="A list of index locations of the new columns.")
    parser.add_argument("--SPLITTING_PATTERN",
                        required=True,
                        help="The regular expression pattern to use to split the column.")
    parser.add_argument("--SEPARATOR",
                        default="\t",
                        type=str,
                        help="The separator.")

    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    PATHS = argNamespace.PATHS
    COLUMN_TO_SPLIT = argNamespace.COLUMN_TO_SPLIT
    NAME_OF_NEW_COLUMNS = argNamespace.NAME_OF_NEW_COLUMNS
    LOCATION_OF_NEW_COLUMNS = argNamespace.LOCATION_OF_NEW_COLUMNS
    SPLITTING_PATTERN = argNamespace.SPLITTING_PATTERN

    SEPARATOR = argNamespace.SEPARATOR

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body
    PARGS = []
    li = sorted(PATHS)
    for fpathString in li:
        fpath = Path(fpathString)
        runOutputPath = runOutputDir.joinpath(fpath.name)
        argsClone = (runOutputPath,
                     fpath,
                     COLUMN_TO_SPLIT,
                     NAME_OF_NEW_COLUMNS,
                     LOCATION_OF_NEW_COLUMNS,
                     SPLITTING_PATTERN,
                     logger,
                     SEPARATOR)
        PARGS.append(argsClone)

    with pp._ProcessPool() as pool:
        results = pool.starmap(expandColumnWrapper, PARGS)

    logger.info(f"""Results are in "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}.""")

    # End module body
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""
Joins two local tables
"""

import argparse
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi import __version__ as drapiVersion
from drapi import loggingChoices
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)

if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--LEFT_TABLE",
                        type=Path,
                        help="The path to the left table to join.",
                        required=True)
    parser.add_argument("--RIGHT_TABLE",
                        help="The path to the right table to join.",
                        required=True)

    parser.add_argument("--LEFT_TABLE_INDEX",
                        default=None,
                        type=str,
                        help="The column to use as index to join on for the left table.")
    parser.add_argument("--RIGHT_TABLE_INDEX",
                        default=None,
                        type=str,
                        help="The column to use as index to join on for the right table.")
    
    if False:
        parser.add_argument("--COLUMNS_TO_JOIN_ON",
                            nargs="+",
                            required=True)

    parser.add_argument("--HOW",
                        type=str,
                        choices=["inner","outer","left","right"],
                        help="The method to join on.")

    parser.add_argument("--COLUMNS_TO_KEEP",
                        default=None,
                        nargs="*",
                        action="extend",
                        help="The columns to keep. By default, all are kept.")
    parser.add_argument("--COLUMNS_TO_DROP",
                        default=[],
                        nargs="*",
                        help="The columns to drop. By defautl, all are kept")
    
    parser.add_argument("--FILE_NAME",
                        type=str,
                        default="Joined Table",
                        help="The name to give to the joined table file.")
    
    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    LEFT_TABLE = argNamespace.LEFT_TABLE
    RIGHT_TABLE = argNamespace.RIGHT_TABLE

    LEFT_TABLE_INDEX = argNamespace.LEFT_TABLE_INDEX
    RIGHT_TABLE_INDEX = argNamespace.RIGHT_TABLE_INDEX

    if False:
        COLUMNS_TO_JOIN_ON = argNamespace.COLUMNS_TO_JOIN_ON

    HOW = argNamespace.HOW

    COLUMNS_TO_KEEP = argNamespace.COLUMNS_TO_KEEP
    COLUMNS_TO_DROP = argNamespace.COLUMNS_TO_DROP

    FILE_NAME = argNamespace.FILE_NAME

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    pass
    # <<< Custom argument parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    pass

    # Check columns to join on
    if False:
        if len(COLUMNS_TO_JOIN_ON) == 1:
            columnsToJoinOn = COLUMNS_TO_JOIN_ON[0]
        elif len(COLUMNS_TO_JOIN_ON) >= 1:
            columnsToJoinOn = COLUMNS_TO_JOIN_ON
        else:
            message = f"""You can only supply either one or two columns to join on."""
            parser.error(message)

    # Check columns to drop and keep
    intersection = set(COLUMNS_TO_DROP).intersection(set(COLUMNS_TO_KEEP))
    if len(intersection) > 0:
        intersectionAsString = ", ".join(sorted(list(intersection)))
        message = f"""The following column names were in both the list of columns to drop and keep: {intersectionAsString}"""
        parser.error(message)
    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    # Load tables
    leftTable = pd.read_csv(LEFT_TABLE, engine="pyarrow")
    rightTable = pd.read_csv(RIGHT_TABLE, engine="pyarrow")

    # Set indices
    indexCounter = 0
    if LEFT_TABLE_INDEX:
        leftTable = leftTable.set_index(LEFT_TABLE_INDEX)
        indexCounter += 1
    
    if RIGHT_TABLE_INDEX:
        rightTable = rightTable.set_index(RIGHT_TABLE_INDEX)
        indexCounter += 1

    joinedTable = leftTable.join(other=rightTable,
                                 how=HOW,
                                 lsuffix="_L",
                                 rsuffix="_R")
    
    # Reset indices with names
    if LEFT_TABLE_INDEX and RIGHT_TABLE_INDEX:
        joinedTable = joinedTable.reset_index(names=LEFT_TABLE_INDEX)
    elif LEFT_TABLE_INDEX:
        joinedTable = joinedTable.reset_index(names=LEFT_TABLE_INDEX)
    elif RIGHT_TABLE_INDEX:
        joinedTable = joinedTable.reset_index(names=RIGHT_TABLE_INDEX)

    
    # Select columns
    columns0 = joinedTable.columns

    # Seelct columns: Columns to drop
    finalColumns = columns0.drop(labels=COLUMNS_TO_DROP)

    # Select columns: Columns to 
    if COLUMNS_TO_KEEP:
        finalColumns = finalColumns[finalColumns.isin(COLUMNS_TO_KEEP)]

    # Select columns: final columns
    finalTable = joinedTable[finalColumns]

    # Make export path
    exportPath = runOutputDir.joinpath(f"""{FILE_NAME}""")
    
    # Save results
    finalTable.to_csv(path_or_buf=exportPath,
                      index=False)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # <<< End script body <<<
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""
Module that contains the inner function for getting data using SQL queries.
"""

from logging import Logger
from pathlib import Path
from typing_extensions import Union
# Third-party packages
import pandas as pd
from sqlalchemy import create_engine
# Local packages
from drapi.code.drapi.classes import SecretString


def getData_inner(connectionString1: str,
                  logger: Logger,
                  queryChunkSize: int,
                  runOutputDir: Path,
                  sqlQuery: Union[None, str],
                  downloadData: bool,
                  connectionString2: SecretString = None,
                  newSQLTable_Name: str = None,
                  newSQLTable_Schema: str = None,
                  outputName: str = None,
                  itstring1: str = None,
                  numChunks1: int = None,
                  sqlFilePath: Union[None, Path, str] = None) -> pd.DataFrame:
    """
    Executes a SQL query.
    """
    connection1 = create_engine(connectionString1).connect().execution_options(stream_results=True)
    connection2 = create_engine(connectionString1).connect().execution_options(stream_results=True)

    # >>> Determine query source: File or string >>>
    # Case 1 (file): `sqlFilePath` is defined
    # Case 2 (string): `sqlQuery` is defined

    if isinstance(sqlFilePath, type(None)):
        case1 = False
    elif isinstance(sqlFilePath, (Path, str)):
        case1 = True
    else:
        message = f"""The variable `sqlFilePath` is of an unexpected type: "{type(sqlFilePath)}"."""
        logger.fatal(message)
        raise Exception(message)

    if isinstance(sqlQuery, type(None)):
        case2 = False
    elif isinstance(sqlQuery, str):
        case2 = True
    else:
        message = f"""The variable `sqlQuery` is of an unexpected type: "{type(sqlQuery)}"."""
        logger.fatal(message)
        raise Exception(message)
    # <<< Determine query source: File or string <<<

    # >>> Define `query` >>>
    if case1 and case2:
        message = f"""There is an ambiguous argument input. Only one of `sqlFilePath` or `sqlQuery` may be passed, but not both."""
        logger.fatal(message)
        raise Exception(message)
    elif case1:
        with open(sqlFilePath, "r") as file:
            query = file.read()
    elif case2:
        query = sqlQuery
    else:
        message = f"""An unexpected error occurred."""
        logger.fatal(message)
        raise Exception(message)
    # <<< Define `query` <<<
    
    # >>> Argumnet check: `outputName` >>>
    if downloadData:
        if isinstance(outputName, type(None)):
            message = """When `downloadData` is set to `True`, you must pass a value to `outputName`."""
            logger.critical(message)
            raise Exception(message)
    else:
        pass

    # <<< Argumnet check: `outputName` <<<

    # Save query to log
    logger.debug(query)

    # Execute query
    logger.info("""  ..  Executing query.""")

    if downloadData:
        logger.info(f"""  ..  Counting the number of query result chunks that are expected with `queryChunkSize` "{queryChunkSize:,}".""")
        queryGenerator0 = pd.read_sql(sql=query, con=connection1, chunksize=queryChunkSize)
        chunks2 = [1 for _ in queryGenerator0]
        numChunks2 = sum(chunks2)
        padlen2 = len(str(numChunks2))
        logger.info(f"""  ..  Counting the number of query result chunks that are expected with `queryChunkSize` "{queryChunkSize:,}" - Done. there are {numChunks2:,} chunks.""")
    else:
        pass

    logger.info("""  ..  Creating query generator.""")
    queryGenerator1 = pd.read_sql(sql=query, con=connection2, chunksize=queryChunkSize)
    logger.info("""  ..  Creating query generator - Done.""")

    logger.info("""  ..  Iterating over query generator.""")
    for it2, queryChunk in enumerate(queryGenerator1, start=1):
        if downloadData:
            logger.info(f"""  ..  ..  Executing query chunk {it2:,} of {numChunks2:,}.""")
            result = queryChunk
            result = pd.DataFrame(result)  # For type hinting
            logger.info("""  ..  ..  Finished query chunk.""")
            logger.info("  ..  ..  Saving chunk.")
            itstring2 = str(it2).zfill(padlen2)
            if itstring1 and numChunks1:
                fpath = runOutputDir.joinpath(f"{outputName} - {itstring1} of {numChunks1} - {itstring2} of {numChunks2}.CSV")
            else:
                fpath = runOutputDir.joinpath(f"{outputName} - {itstring2} of {numChunks2}.CSV")
            result.to_csv(fpath, index=False)
            logger.info("  ..  ..  Saving chunk - done.")
        else:
            logger.info(f"""  ..  ..  Executing query chunk {it2:,}.""")
            result = queryChunk
            result = pd.DataFrame(result)  # For type hinting
            result.to_sql(name=newSQLTable_Name,
                          schema=newSQLTable_Schema,
                          con=connectionString2,
                          if_exists="append",
                          index=False)
            logger.info("""  ..  ..  Finished query chunk.""")

    connection1.close()
    connection2.close()



#!/usr/bin/env python

"""
Get the set of IDs from a Chocolate clinical text workflow.
"""

import argparse
import json
import logging
import os
import pprint
from pathlib import Path
# Third-party packages
import pandas as pd
from sqlalchemy import URL
# Local packages
from drapi import __version__ as drapiVersion
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    makeDirPath)

if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--FILE_PATHS",
                        nargs="*",
                        help="The list of file paths to the tables containing IDs")
    parser.add_argument("--ID_NAMES",
                        nargs="*",
                        type=json.loads,
                        help="A JSON-formatted list of variable name (ID name) pairs to make sets of.")
    parser.add_argument("--ALIASES",
                        nargs="*",
                        type=json.loads,
                        help="""A JSON-formatted list of dictionaries, one dictionary for each `ID_NAME` pair. The dictinoaries are used to rename their corresponding name pairs. If you wish to not rename a variable, map it to the empty string like this: {"original_name": ""}.""")
    parser.add_argument("--ROOT_PATH",
                        help="The root folder for `--FILE_PATHS`")

    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    FILE_PATHS = argNamespace.FILE_PATHS
    ID_NAMES = argNamespace.ID_NAMES
    ALIASES = argNamespace.ALIASES
    ROOT_PATH = argNamespace.ROOT_PATH

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    pass
    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        userID = USER_ID[:]
    else:
        userID = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPwd = os.environ["HFA_UFADPWD"]
    connectionString = URL.create(drivername="mssql+pymssql",
                                  username=userID,
                                  password=userPwd,
                                  host=SERVER,
                                  database=DATABASE)

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    it1Total = len(ID_NAMES)
    nameZip = zip(ID_NAMES, ALIASES)
    # modeDict = {fpath: {"mode": "a",
    #                     "header": True} for fpath in FILE_PATHS}
    for it1, tu in enumerate(nameZip, start=1):
        variableNamePair, aliasDict0 = tu
        aliasDict = {variableAlias: (variableName if variableName != "" else variableAlias) for variableAlias, variableName in aliasDict0.items()}
        logger.info(f"""  Working on variable pair {it1:,} of {it1Total:,}: "{variableNamePair}".""")
        variableNamePair = list(variableNamePair)
        it2Total = len(FILE_PATHS)
        for it2, fpath in enumerate(FILE_PATHS, start=1):
            logger.info(f"""    Working on fpath {it2:,} of {it2Total:,}: "{fpath}".""")
            fpath = Path(fpath)
            df0 = pd.read_csv(filepath_or_buffer=fpath,
                            engine="pyarrow")
            df = df0[variableNamePair].drop_duplicates()

            # Rename variables
            df = df.rename(mapper=aliasDict)

            relativePath = fpath.relative_to(ROOT_PATH)
            variableAlias, variableName = variableNamePair
            variableNameStandard = aliasDict[variableName]
            savePath = runOutputDir.joinpath("Sets",
                                             variableNameStandard,
                                             relativePath.parents[0],
                                             f"{fpath.stem}.CSV")
            savePath.parent.mkdir(parents=True)

            df.to_csv(path_or_buf=savePath,
                                index=False)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # <<< End script body <<<
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""
Compare two cohort (group) files.
"""

import logging
import os
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
from drapi.code.drapi.compareGroups import (compareGroups,
                                            determineMapType,
                                            determineMapTypeFromMap,
                                            mappingAnalysis)

# Arguments
COHORT_1_FILE_PATH = r""
COHORT_2_FILE_PATH = r""
PATIENT_MAP_FILE_PATH = r""
FROM_COLUMN = ""
TO_COLUMN = ""

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 3
IRB_DIR_DEPTH = PROJECT_DIR_DEPTH + 3
IDR_DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 6

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
# ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
#  "IRB_DIRECTORY",                 # noqa
#  "DATA_REQUEST_DIRECTORY",        # noqa
#  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Variables: Other
pass

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `COHORT_1_FILE_PATH`: "{COHORT_1_FILE_PATH}"
    `COHORT_2_FILE_PATH`: "{COHORT_2_FILE_PATH}"
    `PATIENT_MAP_FILE_PATH`: "{PATIENT_MAP_FILE_PATH}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load raw data
    group1data0 = pd.read_csv(COHORT_1_FILE_PATH)
    group2data = pd.read_csv(COHORT_2_FILE_PATH)

    patientMap0 = pd.read_csv(PATIENT_MAP_FILE_PATH)

    # Process data
    patientMap = patientMap0[[FROM_COLUMN, TO_COLUMN]]

    # Mapping analysis
    mappingAnalysis(x0=group1data0,
                    m0=patientMap,
                    logger=logger)
    group1data1_inner = group1data0.dropna().drop_duplicates().set_index(FROM_COLUMN).join(other=patientMap.set_index(FROM_COLUMN),
                                                                                           how="inner",
                                                                                           lsuffix="_L",
                                                                                           rsuffix="_R").reset_index()
    group1data1_outer = group1data0.dropna().drop_duplicates().set_index(FROM_COLUMN).join(other=patientMap.set_index(FROM_COLUMN),
                                                                                           how="outer",
                                                                                           lsuffix="_L",
                                                                                           rsuffix="_R").reset_index()
    group1series1 = group1data1_inner[TO_COLUMN].dropna().drop_duplicates()
    group1series2 = group1data1_outer[TO_COLUMN].dropna().drop_duplicates()
    group2series = group2data[TO_COLUMN].dropna().drop_duplicates()
    compareGroups(group1=group1series1,
                  group2=group2series,
                  logger=logger)
    compareGroups(group1=group1series2,
                  group2=group2series,
                  logger=logger)

    mapType1 = determineMapType(x0=group1data0,
                                x1=group1data1_inner[TO_COLUMN])
    mapType2 = determineMapType(x0=group1data0,
                                x1=group1data1_outer[TO_COLUMN])
    mapType3 = determineMapTypeFromMap(x0=group1data0,
                                       m0=patientMap,
                                       logger=logger)
    logger.info(f"""The mapping of group1 (inner-joined) is of type "{mapType1}".""")
    logger.info(f"""The mapping of group1 (outer-joined) is of type "{mapType2}".""")
    logger.info(f"""The mapping of group1 (from `determineMapTypeFromMap`) is of type "{mapType3}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
De-identifies those above a certain age threshold by replacing the age with a label and setting the birthdate to "1/1/1800". The label is "> X", where "X" is the age threshold.
"""

import logging
import os
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)
# Local packages: Script parameters: General
from drapi.code.drapi.constants.phiVariables import (DICT_OF_PHI_VARIABLES_AGE,
                                          DICT_OF_PHI_VARIABLES_BIRTHDATES)
# Local packages: Script parameters: Paths
# Local packages: Script parameters: File criteria
from common import BO_PORTION_FILE_CRITERIA

# Arguments
AGE_MAP = Path(r"..\Concatenated Results\data\output\makeAgeMap\2024-01-03 13-20-23\Age Map - De-identified Patient Key.CSV")
USE_DE_IDENTIFIED_DATES = True

DE_IDENTIFIED_AGE_VALUE = "> 90"
DE_IDENTIFIED_BIRTHDATE_VALUE = "1/1/1800"

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Variables: Script parameters: Date variables
# NOTE Set `USE_DE_IDENTIFIED_DATES` to `True` if the dates have been de-identified. If the IDs have been de-identified, it is not necessary to set this to `True`.
if USE_DE_IDENTIFIED_DATES:
    DE_IDENTIFICATION_PREFIX = "De-identified "
    dictOfPHIVariablesBirthdates = {portion: [] for portion in DICT_OF_PHI_VARIABLES_BIRTHDATES.keys()}
    for portion, li in DICT_OF_PHI_VARIABLES_BIRTHDATES.items():
        for columnName in li:
            dictOfPHIVariablesBirthdates[portion].append(f"{DE_IDENTIFICATION_PREFIX} {columnName}")
else:
    dictOfPHIVariablesBirthdates = DICT_OF_PHI_VARIABLES_BIRTHDATES

# Variables: Script parameters: Age variables
dictOfPHIVariablesAge = DICT_OF_PHI_VARIABLES_AGE

# Arguments: Portion Paths and conditions
MAC_PATHS = [Path(r"..\Concatenated Results\data\output\deIdentify\...")]  # TODO
WIN_PATHS = [Path(r"..\Concatenated Results\data\output\deIdentify\...")]  # TODO

listOfAgeColumns = [el for value in dictOfPHIVariablesAge.values() for el in value]
listOfBirthdateColumns = [el for value in dictOfPHIVariablesBirthdates.values() for el in value]
listOfColumnsToDeIdentify = listOfAgeColumns + listOfBirthdateColumns
LIST_OF_PORTION_CONDITIONS = [BO_PORTION_FILE_CRITERIA]

# Arguments; General
CHUNK_SIZE = 50000

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "IRB_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Arguments: SQL connection settings
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE = "DWS_PROD"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Variables: Path construction: Project-specific
pass

# Variables: SQL Parameters
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
conStr = f"mssql+pymssql://{uid}:{PWD}@{SERVER}/{DATABASE}"

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `AGE_MAP`: "{AGE_MAP}"
    `USE_MODIFIED_OMOP_DATA_SET`: "{USE_MODIFIED_OMOP_DATA_SET}"
    `MAC_PATHS`: "{MAC_PATHS}"
    `WIN_PATHS`: "{WIN_PATHS}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USERDOMAIN` = "{USERDOMAIN}"
    `USERNAME` = "{USERNAME}"
    `UID` = "{UID}"
    `PWD` = censored
    """)

    # Load age de-identification map
    logger.info("""Loading age de-identification map.""")
    map_ = pd.read_csv(AGE_MAP, comment="#")
    logger.info("""Loading age de-identification map - done.""")
    PATIENT_LOOK_UP_ID_NAME = map_.columns[0]
    mapMask = map_["Over Age Threshold"].apply(lambda integer: True if integer else (False if not integer else None))
    patientsToDeIdentify = map_[PATIENT_LOOK_UP_ID_NAME][mapMask]

    # De-identify by age
    logger.info("""De-identifying files.""")
    columnDeIdentificationRenamePrefix = f"{DE_IDENTIFICATION_PREFIX}"
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                logger.info("""  ..  Reading file to count the number of chunks.""")
                numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE, dtype=str)])
                logger.info(f"""  ..  This file has {numChunks} chunks.""")
                dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE, low_memory=False)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logger.info(f"""  ..    Working on column "{columnName}".""")
                        columnShouldBeProcessed = False
                        if columnName in listOfColumnsToDeIdentify:
                            columnShouldBeProcessed = True
                        else:
                            continue
                        if columnShouldBeProcessed:
                            logger.info("""  ..  ..  Column must be de-identified.""")
                            mask = dfChunk[PATIENT_LOOK_UP_ID_NAME].isin(patientsToDeIdentify)
                            if columnName in listOfAgeColumns:
                                dfChunk.loc[mask, columnName] = DE_IDENTIFIED_AGE_VALUE
                            elif columnName in listOfBirthdateColumns:
                                dfChunk.loc[mask, columnName] = DE_IDENTIFIED_BIRTHDATE_VALUE
                            if USE_DE_IDENTIFIED_DATES:
                                if columnName in listOfAgeColumns:
                                    columnRenamePrefix = columnDeIdentificationRenamePrefix
                                else:
                                    columnRenamePrefix = ""
                            else:
                                columnRenamePrefix = columnDeIdentificationRenamePrefix
                            dfChunk = dfChunk.rename(columns={columnName: f"{columnRenamePrefix}{columnName}"})
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""    This file does not need to be processed.""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.relative_to(projectDir)}".""")



"""
Check the C2S status of MRNs. This checks C2S enrollment and death information.
"""

import argparse
import sys
# Third-party packages
import pandas as pd
from pathlib import Path
# Local packages
from drapi.code.drapi.c2s.c2s import checkStatus, doCheck


def checkStatusWrapper(fpath: str, columnName: str, statusType: str, location: str, verbosity):
    """
    """
    fpath
    data = pd.read_csv(fpath)

    MRNsAsList = data[columnName].to_list()

    result = checkStatus(statusType=statusType,
                         location=location,
                         listOfMRNs=MRNsAsList)

    checkResultPass, failedRows = doCheck(result, statusType)

    if checkResultPass:
        print("Passed.")
    elif not checkResultPass:
        print("Not passed. Below are the failed cases:")
        print(failedRows)

    if verbosity < 10:
        print(result.set_index(columnName).sort_index().to_string())


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--verbosity", help="""Increase output verbosity. See "logging" module's log level for valid values.""", type=int, default=10)

    parser.add_argument("fpath", help="The path to the file that contains the MRNs.", type=str)

    parser.add_argument("columnName", help="The name of the column that contains the MRNs.", type=str)

    parser.add_argument("statusType", help="The type of status check.", choices=["C2S", "death"], type=str)

    parser.add_argument("location", help="The MRN location type.", choices=["gnv", "jax"], type=str)

    args = parser.parse_args()

    fpathAsString = args.fpath
    columnName = args.columnName
    statusType = args.statusType
    location = args.location

    verbosity = args.verbosity

    fpath = Path(fpathAsString)

    checkStatusWrapper(fpath, columnName, statusType, location, verbosity)

    if not len(sys.argv) > 1:
        parser.print_usage()



#!/usr/bin/env python

"""
Uploads a pandas-compatible file to a SQL server
"""

import argparse
import logging
import os
import pprint
import shutil
from pathlib import Path
# Third-party packages
pass
# First-party packages
from drapi import __version__ as drapiVersion
from drapi import loggingChoices
from drapi.code.drapi.classes import (SecretString)
from drapi.code.drapi.cli_parsers import parse_string_to_boolean
from drapi.code.drapi.uploadData import uploadData
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
from drapi.code.drapi.uploadData import uploadData


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments
    parser.add_argument("--directory_or_list_of_paths",
                        required=True,
                        nargs="+")
    parser.add_argument("--schema_name",
                        required=True,
                        type=str)
    parser.add_argument("--table_name",
                        required=True,
                        type=str)
    parser.add_argument("--connection_string",
                        type=SecretString)
    parser.add_argument("--lazy_hack_1",
                        default=False,
                        type=parse_string_to_boolean)

    # Arguments: Meta-parameters
    parser.add_argument("--TIMESTAMP",
                        type=str)
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        type=SecretString,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments
    directory_or_list_of_paths = argNamespace.directory_or_list_of_paths
    schema_name = argNamespace.schema_name
    table_name = argNamespace.table_name
    connection_string = argNamespace.connection_string
    lazy_hack_1 = argNamespace.lazy_hack_1

    # Parsed arguments: Meta-parameters
    TIMESTAMP = argNamespace.TIMESTAMP
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    # >>> Custom argument parsing: determine if input is a directory or list of file paths >>>
    # Case 1: One directory
    # Case 2: One or more files
    # Case 3: All other cases
    arguments_as_files = [Path(el).is_file() for el in directory_or_list_of_paths]
    arguments_as_dirs = [Path(el).is_dir() for el in directory_or_list_of_paths]
    if len(arguments_as_dirs) ==1 and all(arguments_as_dirs):
        # case 1
        directory_path = directory_or_list_of_paths[0]
        list_of_paths = sorted(list([file_path for file_path in Path(directory_path).iterdir()]))
    elif all(arguments_as_files):
        # case 2
        list_of_paths = sorted([Path(file_path) for file_path in directory_or_list_of_paths])
    else:
        raise Exception(f"""`directory_or_list_of_paths` should be a single directory or a list of file paths.""")
    # <<< Custom argument parsing: determine if input is a directory or list of file paths <<<
    # <<< Custom argument parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    # >>> Argument checks: Check booleans >>>
    if isinstance(lazy_hack_1, type(None)):
        parser.error("`lazy_hack_1` must be one of {{True, False}}.")
    # <<< Argument checks: Check booleans <<<
    # <<< Argument checks <<<

    # Variables: Path construction: General
    if TIMESTAMP: 
        runTimestamp = TIMESTAMP
    else:
        runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # Begin module body

    _ = uploadData(list_of_paths=list_of_paths,
                   schema_name=schema_name,
                   table_name=table_name,
                   connection_string=connection_string,
                   lazy_hack_1=lazy_hack_1,
                   logger=logger)

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # End module body
    logger.info(f"""Finished running "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")



"""
Outline

    - Load the cohort defined in "deiden_2021-05-01.db"
"""

# Private variables ("dunders")
__all__ = ["patientMapDf"]

# Imports
from pathlib import Path
import os
import sqlite3
import sys
# Local imports
from drapi.drapi import sqlite2df, getTimestamp

# Arguments
DATABASE_FILE_NAME = "deiden_2021-05-01.db"
DATABSE_DIR_WINDOWS = os.path.join("X:\\",
                                   "FTP",
                                   "IDR",
                                   "ANES",
                                   "IRB201600223 - aka R01",
                                   "Deiden_db")
DATABSE_DIR_MAC = os.path.join("/",
                               "Volumes",
                               "FILES",
                               "FTP",
                               "IDR",
                               "ANES",
                               "IRB201600223 - aka R01",
                               "Deiden_db")

# Variables
this_file_path = Path(__file__)
project_dir = this_file_path.absolute().parent.parent
irb_dir = project_dir.parent
input_dir = os.path.join(project_dir, "data",
                                      "input")
output_dir = os.path.join(project_dir, "data",
                                       "output")
sql_dir = os.path.join(project_dir, "sql")
run_timestamp = getTimestamp()

# Connect to SQLite database
if True:
    # If you have connection to the below directory, use the below line.
    operatingSystem = sys.platform
    if operatingSystem == "win32":
        database_dir = DATABSE_DIR_WINDOWS
    elif operatingSystem == "darwin":
        database_dir = DATABSE_DIR_MAC
    else:
        raise Exception("Unsupported operating system")
elif True:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    database_dir = input_dir
database_path = Path(os.path.join(database_dir,
                                  DATABASE_FILE_NAME))

# SQLite connection
print(f"""[{getTimestamp()}] Loading sqlite database from "{database_path}".""")
sqliteConnection = sqlite3.connect(database_path)
cursor = sqliteConnection.cursor()


def testSQLiteConection():
    """Test SQLite connection"""
    print(f"""[{getTimestamp()}] Testing SQLite connection.""")
    query = """SELECT 1"""
    cursor.execute(query)
    test1 = cursor.fetchall()[0][0]
    if test1:
        print(f"""[{getTimestamp()}] SQLite connection successful: "{test1}".""")


def getPatientMapDf():
    """Query patient de-identification map"""
    print(f"""[{getTimestamp()}] Running SQLite query.""")
    query = """SELECT
            *
            FROM
            PatientDeidenMap
            WHERE
            active_ind_y_n = 1;"""
    cursor.execute(query)
    patientMapLi = cursor.fetchall()
    print(f"""[{getTimestamp()}] SQLite query completed.""")
    patientMapDf = sqlite2df(patientMapLi, "PatientDeidenMap", cursor)
    patientMapDf["deiden_id_string"] = patientMapDf["deiden_id"].apply(lambda integer: f"IDEALIST_2021-05-01_{integer}")
    return patientMapDf


def getEncounterMapDf():
    """Query patient de-identification map"""
    return


if __name__ == "__main__":
    print("This file not implemented to run as script.")
else:
    testSQLiteConection()
    patientMapDf = getPatientMapDf()



"""
Copy files to a single destination directory. Optionally this directory can be added to a compressed archive.
"""

import logging
import os
import pprint
import shutil
import zipfile
from pathlib import Path
from typing_extensions import List
# First-party packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    successiveParents)


def gather_files(destination_folder: str,
                 list_of_directories: List[str],
                 list_of_files: List[List[str]],  # NOTE TODO Not implemented yet.
                 list_of_directories_new_names: List[str],
                 logger: logging.Logger,
                 list_of_loose_files: List[str] = [],
                 timestamp: str = getTimestamp(),
                 overwrite_if_exists_archive: bool = False,
                 overwrite_if_exists_file: bool = False,
                 overwrite_if_exists_folder: bool = False,
                 create_merged_folder: bool = False,
                 create_compressed_archive: bool = True,
                 delete_folder_after_archiving: bool = False):
    """
    Copy files to a single destination directory. Optionally this directory can be added to a compressed archive.
    """
    print(locals())
    arguments_list_as_string = pprint.pformat(locals())
    logger.info(f""" >>> List of arguments >>>
{arguments_list_as_string}
<<< List of arguments <<<""")
    # Define the destination path
    destinationRootFolder = Path(destination_folder)

    # Define archive root folder name NOTE This is useful because if you use the default behavior of `unzip` it would unzip all the portions individually in the current working directory. By using this archive root folder, if the user uses the default behavior, all the portions, including any loose files, are extracted into a single directory.
    archive_root_folder_name = f"IDR Data Release - {timestamp}"

    # Copy files and directories
    if create_merged_folder:
        if destinationRootFolder.exists():
            logger.warning(f"""WARNING: The destination folder already exists: "{destinationRootFolder.absolute()}".""")
            if overwrite_if_exists_folder:
                logger.info("""  Removing folder contents to make room for new files.""")
                for fpath in destinationRootFolder.iterdir():
                    logger.info(f"""    Removing "{fpath.absolute()}".""")
                    os.remove(fpath)
                    logger.info(f"""    Removing "{fpath.absolute()}" - done.""")
            else:
                msg = "  The destination folder exists and no option was passed to over-write it."
                logger.critical(msg)
                raise Exception(msg)
        else:
            logger.info(f"""Making destination folder: "{destinationRootFolder.absolute()}".""")
            makeDirPath(destinationRootFolder)

        logger.info("""Working on list of files.""")
        for fpathString in list_of_loose_files:
            fpath = Path(fpathString)
            logger.info(f"""  Working on file "{fpath.absolute()}".""")
            dest = destinationRootFolder.joinpath(fpath.name)
            logger.info(f"""    Saving to "{dest.absolute()}".""")
            shutil.copyfile(fpath, dest)
        logger.info("""Working on list of files - done.""")

        logger.info("""Working on list of directories.""")
        if list_of_directories_new_names:
            listOfDirectoriesNewNames = list_of_directories_new_names[:]
        else:
            listOfDirectoriesNewNames = ["" for _ in list_of_directories]
        directoryPathsAndNames = zip(list_of_directories, listOfDirectoriesNewNames)
        for directoryString, newDirectoryName in directoryPathsAndNames:
            directory = Path(directoryString)
            destinationFolder = destinationRootFolder.joinpath(newDirectoryName)
            destinationFolder.mkdir()
            listOfFiles1 = sorted(list(directory.iterdir()))
            for fpath in listOfFiles1:
                logger.info(f"""  Working on file "{fpath.absolute()}".""")
                dest = destinationFolder.joinpath(fpath.name)
                logger.info(f"""    The destination path is "{dest.absolute()}".""")
                if dest.exists():
                    msg = f"""    WARNING: This file already exists: "{dest}"."""
                    logger.warning(msg)
                    if overwrite_if_exists_file:
                        continueOperation = True
                    else:
                        continueOperation = False
                else:
                    continueOperation = True
                if continueOperation:
                    logger.info("""    Saving to destination path.""")
                    shutil.copyfile(fpath, dest)
                else:
                    logger.info("""    The file was not saved to the destination path. File over-write is set to `False`.""")
        logger.info("""Working on list of directories - done.""")

    # Create compressed archive
    if create_compressed_archive:
        archivePath = destinationRootFolder.with_suffix(".ZIP")
        logger.info(f"""Creating compressed archive: "{archivePath.absolute()}"".""")
        archivePathParentDir, _ = successiveParents(pathObj=archivePath,
                                    numLevels=1)
        if overwrite_if_exists_archive:
            if archivePath.exists():
                logger.info("""The archive folder already exists and will be removed before writing.""")
                os.remove(archivePath)

        # Define list of files to archive, and their paths
        if list_of_directories_new_names:
            listOfDirectoriesNewNames = list_of_directories_new_names[:]
        else:
            listOfDirectoriesNewNames = ["" for _ in list_of_directories]
        directoryPathsAndNames = zip(list_of_directories, listOfDirectoriesNewNames)
        listOfFiles2All = []
        for directoryString, newDirectoryName in directoryPathsAndNames:
            directory = Path(directoryString)
            listOfFiles2 = sorted(list(directory.iterdir()))
            for fpath2 in listOfFiles2:
                dest2 = Path(archive_root_folder_name).joinpath(newDirectoryName, fpath2.name)
                listOfFiles2All.append((fpath2, dest2))
        
        logger.info(f"""These are the files paths for the archived items:\n{pprint.pformat(listOfFiles2All)}.""")

        with zipfile.ZipFile(file=archivePath,
                             mode="a",
                             compression=zipfile.ZIP_DEFLATED) as zipObj:
            logger.info("Adding files to archive.")
            for fpath2, dest2 in sorted(listOfFiles2All):
                logger.info(f"""  Working on file "{fpath2.absolute()}".""")
                logger.info(f"""    This has a destination path of "{dest2}".""")
                zipObj.write(filename=fpath2, arcname=dest2)
        logger.info("Creating compressed archive - done.")

        if create_merged_folder:
            if delete_folder_after_archiving:
                logger.info("""Removing intermediate folder.""")
                shutil.rmtree(destinationRootFolder)
                logger.info("""Removing intermediate folder - done.""")
            else:
                pass



"""
Testing parsing command-line string representation of bytes.

1. The user types bytes using string literal notation in the command line
2. Arguments go through argparse and a custom parser to convert the strings into bytes.
"""

import argparse
import ast
from typing_extensions import Dict

from drapi.code.drapi.cli_parsers import parse_string_to_bytes

# Script arguments
COMMAND_LINE_INPUT = """--arg_string \\x00\
                        --arg_bytes \\x00\
                        """
VALUES_TO_COMPARE = {1: {"name": "null byte",
                         "value": b"\x00"},
                     2: {"name": "escaped null byte",
                         "value": "\\x00"}}


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--arg_string",
                        type=str)
    parser.add_argument("--arg_bytes",
                        type=parse_string_to_bytes)

    argNamespace = parser.parse_args(COMMAND_LINE_INPUT.split())

    arg_string: str = argNamespace.arg_string
    arg_bytes: bytes = argNamespace.arg_bytes

    feedback_message = f"""\
`arg_string` type:  {type(arg_string)}
`arg_string` value: {arg_string}
`arg_bytes` type:   {type(arg_bytes)}
`arg_bytes` value:  {arg_bytes}\
"""
    print(feedback_message)

    for it, di in VALUES_TO_COMPARE.items():
        di: Dict[str, str] = di
        value_to_compare_name = di["name"]
        value_to_compare = di["value"]
        comparison_message = f"""\
    Is equal to {value_to_compare_name} ({repr(value_to_compare)})?
        `arg_string`: {arg_string == value_to_compare}
        `arg_bytes`:  {arg_bytes == value_to_compare}\
"""
        print(comparison_message)

    print("""Printing "utf-8" characters from 0 to 128.""")
    for integer in range(128):
        bytes_array = bytes([integer])
        bytes_array_repr = repr(bytes_array)
        bytes_array_decoded = bytes_array.decode(encoding="utf-8")
        bytes_array_decoded_repr = repr(bytes_array_decoded)
        bytes_array_escaped = ast.literal_eval(f"r{bytes_array_decoded_repr}")
        bytes_array_escaped_repr = repr(bytes_array_escaped)
        bytes_from_string = parse_string_to_bytes(string_=bytes_array_escaped)
        bytes_from_string_repr = repr(bytes_from_string)
        print(f"Working on integer {integer}:")
        print(f"  Bytes array:\t\t\t{bytes_array_repr}")
        print(f"  Bytes array decoded:\t\t{bytes_array_decoded}")
        print(f"  Bytes array escaped:\t\t {bytes_array_escaped_repr}")
        print(f"  String parsed as bytes:\t{bytes_from_string_repr}")
        if integer == 0:
            special_case_1_result = bytes_array_escaped == "\\x00"
            special_case_2_result = bytes_from_string == b"\x00"
            message_1 = f"""    Special case. Testing if the null byte was correctly escaped. Result: {special_case_1_result}."""
            message_2 = f"""    Special case. Testing if the null byte was correctly escaped and reconstructed. Result: {special_case_2_result}."""
            print(message_1)
            print(message_2)
        print()



"""
Utility functions for the notes pipelines.
"""

import re
from typing import Union


def isNotesChunk(string: str, cohortName: str) -> Union[bool, None]:
    """
    Checks if a filename is the intermediate chunked version of the final table and returns the corresponding boolean. If the file name does not match the notes pipeline filename format, it returns `None`.
    """
    pattern = rf"^(?P<cohortName>{cohortName})_(?P<noteType>note|order_impression|order|order_narrative|order_result_comment)_metadata_?(?P<chunkNum>\d+)?.csv$"
    matchObj = re.match(pattern, string)
    if matchObj:
        groupdict = matchObj.groupdict()
        chunkNum = groupdict["chunkNum"]
        if chunkNum:
            isChunk = True
        else:
            isChunk = False
    else:
        isChunk = None
    return isChunk



"""
Parse variable alias arguments in package scripts.
"""

from typing_extensions import Dict

# Custom argument parsing: aliases


def parseAliasArguments(customAliases: Dict[str, str],
                        useDefaultAliases: bool,
                        defaultAliases: Dict[str, str]) -> Dict[str, str]:
    """
    Parse variable alias arguments in package scripts.
    """
    if customAliases:
        customAliases = dict(customAliases)  # For type hinting
    if customAliases and useDefaultAliases:
        variableAliases = defaultAliases.copy()
        listOfAliasesToOverwrite = sorted(list(set(customAliases.keys()).intersection(set(variableAliases.keys()))))
        if len(listOfAliasesToOverwrite) > 0:
            print(f"""WARNING: The following aliases are being over-written by your custom input:\n{listOfAliasesToOverwrite}.""")
        variableAliases.update(customAliases)
    elif customAliases:
        variableAliases = customAliases.copy()
    elif useDefaultAliases:
        variableAliases = defaultAliases.copy()
    else:
        variableAliases = {}
    return variableAliases



from typing import (Literal,
                    Union)
# Third-party packages
import pandas as pd
from pandas._libs.tslibs import BaseOffset


def makeListOfDates(start: str,
                    end: str,
                    period: Union[Literal["D", "h", "M", "Q", "s", "Y"],
                                  BaseOffset],
                    periodValue: Union[int, float],
                    dateFormat: str= "%Y-%m-%d"):
    """
    Makes a list of dates of the form `'date1' and 'date2', 'date2' and 'date3'`.)
    """
    listOfDates = []
    timeStamp_0 = pd.to_datetime(start)
    timeStamp_n = pd.to_datetime(end)
    period_0 = timeStamp_0.to_period(period)
    period_1 = period_0 + periodValue
    timeStamp_1 = period_1.to_timestamp()
    while timeStamp_1.toordinal() < timeStamp_n.toordinal():
        date_0 = timeStamp_0.strftime(dateFormat)
        date_1 = timeStamp_1.strftime(dateFormat)
        string = f"""'{date_0}' and '{date_1}'"""
        listOfDates.append(string)
        timeStamp_0 = timeStamp_1
        period_0 = timeStamp_0.to_period(period)
        period_1 = period_0 + periodValue
        timeStamp_0 = period_0.to_timestamp()
        timeStamp_1 = period_1.to_timestamp()
    timeStamp_1 = timeStamp_n
    date_0 = timeStamp_0.strftime(dateFormat)
    date_1 = timeStamp_1.strftime(dateFormat)
    string = f"""'{date_0}' and '{date_1}'"""
    listOfDates.append(string)
    return listOfDates




"""
Test command-line argument parsing with Python's `argparse` module.
"""

import argparse
# Third-party packages
pass
# First-party packages
from drapi.code.drapi.cli_parsers import parse_string_to_boolean

if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("--boolean",
                        default=False,
                        type=parse_string_to_boolean)

    argNamespace = parser.parse_args()

    # Parsed arguments: Main: Multiple query option
    boolean = argNamespace.boolean

    print(boolean)



"""
Replaces text using byte strings.

This is useful for converting CRLF to LF. You can do so like this:

```bash
python replaceText.py -f FILEPATH -bo $'\\r\\n' -bn $'\\n'
```
"""

import argparse
import pprint
from pathlib import Path


def previewFile(fpath: Path,
                numCharacters: int=100):
    """
    """
    with open(fpath, "rb") as file:
        text = file.read(numCharacters)
    print(text)


def convertNewlines(fpath: Path,
                    oldLineEnding: bytes,
                    newLineEnding: bytes):
    """
    """
    with open(fpath, "rb") as file:
        text = file.read()
        newText = text.replace(oldLineEnding, newLineEnding)
    with open(fpath, "wb") as file:
        file.write(newText)


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("-d",
                        "--LIST_OF_DIRECTORIES",
                        nargs="*",
                        type=Path,
                        help="List of directories whose contents are to be edited.")
    parser.add_argument("-f",
                        "--LIST_OF_FILES",
                        nargs="*",
                        type=Path,
                        help="List of files to edit.")
    HINT = """HINT, to pass an escaped character like "\\n" in a BASH-like shell, use `$'\\n'`. Note the single quotes."""
    parser.add_argument("-bo",
                        "--BYTES_OLD",
                        type=lambda string: bytes(string, encoding="utf-8"),
                        help=f"""The bytes to replace. {HINT}""")
    parser.add_argument("-bn",
                        "--BYTES_NEW",
                        type=lambda string: bytes(string, encoding="utf-8"),
                        help=f"The bytes to replace with. {HINT}")


    parser.add_argument("-p",
                        "--PREVIEW_FILE",
                        type=Path,
                        help="The path to a file to preview in bytes.")
    parser.add_argument("-nc",
                        "--PREVIEW_FILE_NUM_CHARS",
                        type=int,
                        help="The number of bytes to preview from a file.")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    LIST_OF_DIRECTORIES = argNamespace.LIST_OF_DIRECTORIES
    LIST_OF_FILES = argNamespace.LIST_OF_FILES
    OLD_LINE_ENDING = argNamespace.OLD_LINE_ENDING
    NEW_LINE_ENDING = argNamespace.NEW_LINE_ENDING

    PREVIEW_FILE = argNamespace.PREVIEW_FILE
    PREVIEW_FILE_NUM_CHARS = argNamespace.PREVIEW_FILE_NUM_CHARS
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    if LIST_OF_FILES or LIST_OF_DIRECTORIES:
        pass
    elif PREVIEW_FILE:
        pass
    else:
        parser.error("Although `LIST_OF_FILES`, `LIST_OF_DIRECTORIES`, and `PREVIEW_FILE` are marked as optional in the help text, you must actually provide arguments for at least one of them.")
    # <<< Custom argument parsing <<<

    # >>> Module variables >>>        
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    # <<< Module variables <<<

    # >>> Logging block >>>
    print(f"""Running "{thisFileStem}".""")
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)
    print(f"""Script arguments:\n{argListString}""")
    # <<< Logging block <<<

    # >>> Begin module body >>>
    if LIST_OF_FILES:
        listOfFiles1 = sorted([Path(path) for path in LIST_OF_FILES])
        for fpath in LIST_OF_FILES:
            fpath = Path(fpath)  # For type hinting
            convertNewlines(fpath=fpath,
                            oldLineEnding=OLD_LINE_ENDING,
                            newLineEnding=NEW_LINE_ENDING)
    elif LIST_OF_DIRECTORIES:
        for dpath in LIST_OF_DIRECTORIES:
            dpath = Path(dpath)  # For type hinting
            listOfFiles2 = sorted(list(dpath.iterdir()))
            for fpath in listOfFiles2:
                convertNewlines(fpath=fpath,
                                oldLineEnding=OLD_LINE_ENDING,
                                newLineEnding=NEW_LINE_ENDING)
    elif PREVIEW_FILE:
        previewFile(fpath=PREVIEW_FILE,
                    numCharacters=PREVIEW_FILE_NUM_CHARS)

    print(f"""Finished running "{thisFileStem}".""")


"""
Replaces text using byte strings.

This is useful for converting CRLF to LF. You can do so like this:

```bash
python replace_text.py -list_of_file_paths "$list_of_file_paths" -old_text $'\\r\\n' -new_text $'\\n'
```
"""

import argparse
import logging
import multiprocessing as mp
import os
import pprint
import shutil
from functools import partial
from pathlib import Path
from typing import List
# Third-party packages
pass
# First-party packages
from drapi.code.drapi.cli_parsers import parse_string_to_bytes
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
from drapi.code.drapi.replace_text import replace_text_wrapper
from drapi import loggingChoices
from drapi import __version__ as drapiVersion


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments
    parser.add_argument("--list_of_file_paths",
                        type=Path,
                        required=True,
                        nargs="+",
                        help="The list of file paths to convert.")

    # Arguments: group 1
    parser_group_old_text = parser.add_mutually_exclusive_group(required=True)
    parser_group_old_text.add_argument("--old_text",
                        type=str,
                        help="The text to replace.")
    parser_group_old_text.add_argument("--old_text_bytes",
                        type=parse_string_to_bytes,
                        help="""The text to replace, as Python literal bytes, e.g. `"\\x00"` for `b"\\x00"` or `"Column 1\\tColumn2" for `b"Column 1\\tColumn2".""")

    # Arguments: group 2
    parser_group_new_text = parser.add_mutually_exclusive_group(required=True)
    parser_group_new_text.add_argument("--new_text",
                        type=str,
                        help="The text with which to replace.")
    parser_group_new_text.add_argument("--new_text_bytes",
                        type=parse_string_to_bytes,
                        help="""The text with which to replace, as Python literal bytes, e.g. `"\\x00"` for `b"\\x00"` or `"Column 1\\tColumn2" for `b"Column 1\\tColumn2".""")
    
    # Arguments: Meta-parameters
    parser.add_argument("--TIMESTAMP",
                        type=str)
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    argNamespace = parser.parse_args()

    # Parsed arguments
    list_of_file_paths: List[Path] = argNamespace.list_of_file_paths

    old_text: str = argNamespace.old_text
    new_text: str = argNamespace.new_text
    old_text_bytes: bytes = argNamespace.old_text_bytes
    new_text_bytes: bytes = argNamespace.new_text_bytes

    # Parsed arguments: Meta-parameters
    TIMESTAMP: str = argNamespace.TIMESTAMP
    LOG_LEVEL: str = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    # >>> Custom argument parsing: Parsing 1 >>>
    if True:
        if any([old_text_bytes, new_text_bytes]):
            as_bytes = True
        else:
            as_bytes = False

        if old_text:
            if as_bytes:
                old_text_ = old_text.encode()
            else:
                old_text = old_text
        elif old_text_bytes:
            old_text_ = old_text_bytes
        else:
            raise Exception
        
        if new_text:
            if as_bytes:
                new_text_ = new_text.encode()
            else:
                new_text_ = new_text
        elif new_text_bytes:
            new_text_ = new_text_bytes
        else:
            raise Exception
    # <<< Custom argument parsing: Parsing 1 <<<

    # >>> Custom argument parsing: Parsing 2 >>>
    # <<< Custom argument parsing: Parsing 2 <<<
    # <<< Custom argument parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    # >>> Argument checks: Check 1 >>>
    pass
    # <<< Argument checks: Check 1 <<<
    # >>> Argument checks: Check 2 >>>
    pass
    # <<< Argument checks: Check 2 <<<
    # <<< Argument checks <<<

    # Variables: Path construction: General
    if TIMESTAMP:
        runTimestamp = TIMESTAMP
    else:
        runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    # Parallel implementation
    with mp.Pool() as pool:
        # NOTE TODO For developer: should not pass implicitely-valued None-types to the function. I should handle this better, maybe by using `eval`. See https://www.geeksforgeeks.org/execute-string-code-python/
        results = pool.map(partial(replace_text_wrapper,
                                old_text=old_text_,
                                new_text=new_text_,
                                output_dir=runOutputDir),
                        list_of_file_paths)

    # <<< End script body <<<

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # Script end confirmation
    logger.info(f"""Finished running "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")



"""
De-identify files

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# TODO Assign portion name to each path (per OS) so that portion files are stored in their respective folders, this prevents file from being overwritten in the unlikely, but possible, case files from different portions have the same name.
# TODO Investigate if a symlink can be made for files that are copied without alteration, to save space and time on larger projects.
"""

import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (flatExtend,
                                    getTimestamp,
                                    makeDirPath,
                                    readDataFile,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import (LIST_OF_PHI_VARIABLES_OMOP_UNINFORMATIVE,
                                          LIST_OF_PHI_VARIABLES_OMOP_BIRTHDATE_CONDITIONAL)
# Project parameters: General
from .common import (STUDY_TYPE,
                     DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
# Project parameters: Portion paths and criteria
from .common import (MODIFIED_OMOP_PORTION_DIR_MAC, MODIFIED_OMOP_PORTION_DIR_WIN,
                     CLINICAL_TEXT_PORTION_DIR_MAC, CLINICAL_TEXT_PORTION_DIR_WIN,
                     OMOP_PORTION_DIR_MAC, OMOP_PORTION_DIR_WIN)

# Arguments: Use concatenated files
USE_CONCATENATED_FILES = True  # TODO

# Arguments: OMOP data set selection
USE_MODIFIED_OMOP_DATA_SET = True

# Arguments: File location definition: By concatenation
CONCATENATED_PORTIONS_DIR_MAC = Path("../../Concatenated Results/data/output/convertColumnsHash/.../Portions/OMOP")  # TODO
CONCATENATED_PORTIONS_DIR_WIN = Path("../../Concatenated Results/data/output/convertColumnsHash/.../Portions/OMOP")  # TODO

# Arguments: Portion Paths and conditions
if USE_MODIFIED_OMOP_DATA_SET:
    OMOPPortionDirMac = MODIFIED_OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = MODIFIED_OMOP_PORTION_DIR_WIN
else:
    OMOPPortionDirMac = OMOP_PORTION_DIR_MAC
    OMOPPortionDirWin = OMOP_PORTION_DIR_WIN

# Arguments: File location definition: OS-specific
# There are two typical workflows. Deleting columns in the beginning of the workflow, or towards the end, after it's been de-identified (i.e., "concatenated")

if USE_CONCATENATED_FILES:
    MAC_PATHS = [CONCATENATED_PORTIONS_DIR_MAC]
    WIN_PATHS = [CONCATENATED_PORTIONS_DIR_WIN]
else:
    MAC_PATHS = [CLINICAL_TEXT_PORTION_DIR_MAC,
                 OMOPPortionDirMac]
    WIN_PATHS = [OMOPPortionDirWin,
                 CLINICAL_TEXT_PORTION_DIR_WIN]

# Arguments: Definition of criteria for file release
# NOTE (Developer's Note) The files to release and the file criteiria both act as criteria to release. The argument structure here is not very clear and it will take some time to create a generalizeable template. However, it seems that `LIST_OF_PORTION_CONDITIONS` is the only output of this arguments section, i.e., the only require input for the script. Also note that each portion has its own criteria, but they are not used in the template.
BO_FILES_TO_RELEASE = []  # TODO

NOTES_COHORT_NAME = ""  # TODO
NOTES_METADATA_FILES_TO_RELEASE = ["provider_metadata.csv",
                                   f"{NOTES_COHORT_NAME}_note_metadata.csv",
                                   f"{NOTES_COHORT_NAME}_order_impression_metadata.csv",
                                   f"{NOTES_COHORT_NAME}_order_metadata.csv",
                                   f"{NOTES_COHORT_NAME}_order_narrative_metadata.csv",
                                   f"{NOTES_COHORT_NAME}_order_result_comment_metadata.csv"]

I2B2_COHORT_NAME = ""  # TODO
I2B2_FILES_TO_RELEASE = [f"{I2B2_COHORT_NAME}_observation_fact_GNV.csv",
                         f"{I2B2_COHORT_NAME}_observation_fact_JAX.csv",
                         f"{I2B2_COHORT_NAME}_patient_dimension_GNV.csv",
                         f"{I2B2_COHORT_NAME}_patient_dimension_JAX.csv",
                         f"{I2B2_COHORT_NAME}_visit_dimension_GNV.csv",
                         f"{I2B2_COHORT_NAME}_visit_dimension_JAX.csv"]

OMOP_FILES_TO_RELEASE = ["condition_occurrence.csv",
                         "death.csv",
                         "device_exposure.csv",
                         "drug_exposure.csv",
                         "episode.CSV",
                         "location.csv",
                         "measurement_laboratories.csv",
                         "measurement.csv",
                         "observation_period.csv",
                         "observation.csv",
                         "person.csv",
                         "procedure_occurrence.csv",
                         "visit_occurrence.csv"]

ZIP_CODE_FILES_TO_RELEASE = ["zipcodes.csv"]

FILES_TO_RELEASE = flatExtend([BO_FILES_TO_RELEASE,
                               NOTES_METADATA_FILES_TO_RELEASE,
                               OMOP_FILES_TO_RELEASE])
FILES_TO_RELEASE = [fname.lower() for fname in FILES_TO_RELEASE]


CONCATENATED_PORTIONS_FILE_CRITERIA = [lambda pathObj: pathObj.name.lower() in FILES_TO_RELEASE,
                                       lambda pathObj: pathObj.suffix.lower() == ".csv",
                                       lambda pathObj: pathObj.is_file()]

LIST_OF_PORTION_CONDITIONS = [CONCATENATED_PORTIONS_FILE_CRITERIA]

# Arguments: Columns to delete: By portion
# TODO Columns to delete. The list variable deletes column matches in any file. The dictionary variable delete the columns that are in the file named in the dictionary key.
if STUDY_TYPE == "Non-Human":
    COLUMNS_TO_DELETE_BO = ["HIV Status"]
else:
    COLUMNS_TO_DELETE_BO = []

COLUMNS_TO_DELETE_I2B2 = ["LOCATION_CD"]

COLUMNS_TO_DELETE_OMOP = ["person_source_value"] + LIST_OF_PHI_VARIABLES_OMOP_UNINFORMATIVE

COLUMNS_TO_DELETE = flatExtend([COLUMNS_TO_DELETE_BO,
                                COLUMNS_TO_DELETE_I2B2,
                                COLUMNS_TO_DELETE_OMOP])

# Arguments: Columns to delete: By file
if STUDY_TYPE == "Non-Human":
    COLUMNS_TO_DELETE_OMOP_TABLE_PERSON = LIST_OF_PHI_VARIABLES_OMOP_BIRTHDATE_CONDITIONAL
else:
    COLUMNS_TO_DELETE_OMOP_TABLE_PERSON = []
if STUDY_TYPE == "Non-Human":
    COLUMNS_TO_DELETE_OMOP_TABLE_LOCATION = ["address_1",
                                             "address_2",
                                             "city",
                                             "county",
                                             "latitude",
                                             "longitude",
                                             "zip"]
elif STUDY_TYPE == "Limited Data Set (LDS)":
    COLUMNS_TO_DELETE_OMOP_TABLE_LOCATION = ["address_1",
                                             "address_2",
                                             "latitude",
                                             "longitude"]
else:
    raise Exception(f"""Unexpected value for `STUDY_TYPE`: "{STUDY_TYPE}".""")

COLUMNS_TO_DELETE_DICT = {"condition_occurrence": [],
                          "death": [],
                          "device_exposure": [],
                          "drug_exposure": ["sig"],
                          "encounters": [],
                          "location": COLUMNS_TO_DELETE_OMOP_TABLE_LOCATION,
                          "measurement": [],
                          "measurement_laboratories": [],
                          "observation": [],
                          "observation_period": [],
                          "person": COLUMNS_TO_DELETE_OMOP_TABLE_PERSON,
                          "procedure_occurrence": [],
                          "visit_occurrence": []}

# Arguments: General

CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "linux":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")

    # De-identify columns
    logging.info("""Deleting columns not authorized for release.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logging.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        listOfFiles = sorted(list(directory.iterdir()))
        for file in listOfFiles:
            logging.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            logging.info(f"""  Conditions: "{conditions}".""")
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logging.info("""    File has met all conditions for processing.""")
                chunkGenerator1 = readDataFile(file, chunksize=CHUNK_SIZE, low_memory=False)
                chunkGenerator2 = readDataFile(file, chunksize=CHUNK_SIZE, low_memory=False)
                logging.info(f"""  ..  Reading file to count the number of chunks.""")
                it1Total = sum([1 for _ in chunkGenerator1])
                logging.info(f"""  ..  Reading file to count the number of chunks. There are {it1Total:,} chunks.""")
                for it1, dfChunk in enumerate(chunkGenerator2, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logging.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logging.info(f"""  ..    Working on column "{columnName}".""")
                        if file.stem in COLUMNS_TO_DELETE_DICT.keys():
                            listOfColumns = COLUMNS_TO_DELETE + COLUMNS_TO_DELETE_DICT[file.stem]
                        else:
                            listOfColumns = COLUMNS_TO_DELETE
                        if columnName in listOfColumns:
                            logging.info("""  ..  ..  Column must be deleted. Deleting column.""")
                            dfChunk = dfChunk.drop(columns=columnName)
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logging.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logging.info("""    This file does not need to be processed.""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Makes de-identification maps, building on existing maps.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects integer IDs, so no string IDs like Epic Patient IDs.
"""

import argparse
import logging
import os
import pprint
import json
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
from sqlalchemy import URL
# Local packages
from drapi import __version__ as drapiVersion
from drapi.code.drapi.drapi import (choosePathToLog,
                                    ditchFloat,
                                    getTimestamp,
                                    handleDatetimeForJson,
                                    makeDirPath,
                                    makeMap,
                                    makeSetComplement,
                                    successiveParents)
from drapi.code.drapi.constants.phiVariables import (FILE_NAME_TO_VARIABLE_NAME_DICT,
                                                     VARIABLE_NAME_TO_FILE_NAME_DICT)


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("--SETS_PATH",
                        required=True,
                        type=Path,
                        help="The path to the directory that contains the variable set of values for which to create maps.")
    
    # Arguments: Common to the de-identification suite
    parser.add_argument("--IRB_NUMBER")
    parser.add_argument("--CUSTOM_ALIASES",
                        help="""A JSON-formatted string of the form {`ALIAS`: `VARIABLE_NAME`}, where `VARIABLE_NAME` is the BO version of a variable name, and `ALIAS` is an alias of the variable name. An example is {"EncounterCSN": "Encounter # (CSN)"}.""",
                        type=json.loads)

    # Arguments: 
    parser.add_argument("--DEFAULT_ALIASES",
                        help="""Indicates whether to include the default IDR aliases for variables.""",
                        action="store_true")

    # Arguments: Other
    parser.add_argument("--PANDAS_ENGINE",
                        help="""The pandas engine to use when reading data files.""",
                        default="pyarrow",
                        choices=["c",
                                 "none",
                                 "python",
                                 "pyarrow"])

    # Arguments: Meta-parameters
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    pass

    # Parsed arguments: Meta-parameters
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    pass
    # <<< Argument checks <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        userID = USER_ID[:]
    else:
        userID = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPwd = os.environ["HFA_UFADPWD"]
    connectionString = URL.create(drivername="mssql+pymssql",
                                  username=userID,
                                  password=userPwd,
                                  host=SERVER,
                                  database=DATABASE)

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    # Get set of values
    # NOTE The code that used to be in this section was moved to "getIDValues.py"
    logging.info(f"""Using the set of new values in the directory "{getIDValuesOutput.absolute().relative_to(rootDirectory)}".""")

    # QA: Make sure all data variables are present in the script parameters.
    collectedVariables = [FILE_NAME_TO_VARIABLE_NAME_DICT[fname.stem] for fname in getIDValuesOutput.iterdir()]
    missingDataTypes = []
    missingVariableSuffixes = []
    for variableName in collectedVariables:
        if variableName not in DATA_TYPES_DICT.keys():
            missingDataTypes.append(variableName)
        if variableName not in VARIABLE_SUFFIXES.keys():
            missingVariableSuffixes.append(variableName)
    if len(missingDataTypes) > 0:
        text = "\n".join(missingDataTypes)
        raise Exception(f"""Not all variables have a data type assigned to them:\n{text}""")
    if len(missingVariableSuffixes) > 0:
        text = "\n".join(missingVariableSuffixes)
        raise Exception(f"Not all variables have a de-identification suffix assigned to them:\n{text}")

    # Create reverse-look-up alias map
    variableAliasesReverse = {}
    for variableAlias, variableMainName in VARIABLE_ALIASES.items():
        if variableMainName in variableAliasesReverse.keys():
            variableAliasesReverse[variableMainName].append(variableAlias)
        else:
            variableAliasesReverse[variableMainName] = [variableAlias]

    # Concatenate all old maps
    oldMaps = {}
    logging.info("""Concatenating all similar pre-existing maps.""")
    for variableName in collectedVariables:
        logging.info(f"""  Working on variable "{variableName}".""")
        # Get maps explicitely or implicietly referring to this variable name.
        condition1 = variableName in OLD_MAPS_DIR_PATH.keys()
        if condition1:
            variableLookupName = variableName
            logging.info("""    Variable has pre-existing map(s).""")
            listOfMapPaths = OLD_MAPS_DIR_PATH[variableLookupName]
            dfConcat = pd.DataFrame()
            for mapPath in listOfMapPaths:
                logging.info(f"""  ..  Reading pre-existing map from "{mapPath.absolute().relative_to(rootDirectory)}".""")
                df = pd.DataFrame(pd.read_csv(mapPath))
                dfConcat = pd.concat([dfConcat, df])
            oldMaps[variableName] = dfConcat
        if variableName in variableAliasesReverse.keys():
            for variableAlias in variableAliasesReverse[variableName]:
                condition2 = variableAlias in OLD_MAPS_DIR_PATH.keys()
                if condition2:
                    variableLookupName = variableAlias
                    logging.info("""    Variable has pre-existing aliased map(s).""")
                    listOfMapPaths = OLD_MAPS_DIR_PATH[variableLookupName]
                    dfConcat = pd.DataFrame()
                    for mapPath in listOfMapPaths:
                        logging.info(f"""  ..  Reading pre-existing map from "{mapPath}".""")
                        df = pd.DataFrame(pd.read_csv(mapPath))
                        dftemp = makeMap(IDset=set(), IDName=variableName, startFrom=0, irbNumber=IRB_NUMBER, suffix="", columnSuffix=variableName, deIdentificationMapStyle="lemur", logger=logging.getLogger())
                        df.columns = dftemp.columns
                        dfConcat = pd.concat([dfConcat, df])
                    if condition1:
                        df = oldMaps[variableName]
                        dfConcat = pd.concat([dfConcat, df])
                        oldMaps[variableName] = dfConcat
                    else:
                        oldMaps[variableName] = dfConcat
        else:
            condition2 = False
        if not (condition1 or condition2):
            logging.info("""    Variable has no pre-existing map.""")
            oldMaps[variableName] = pd.DataFrame()

    # Get the set difference between all old maps and the set of un-mapped values
    valuesToMap = {}
    setsToMapDataDir = runIntermediateDataDir.joinpath("valuesToMap")
    makeDirPath(setsToMapDataDir)
    logging.info("""Getting the set difference between all old maps and the set of un-mapped values.""")
    for variableName in collectedVariables:
        logging.info(f"""  Working on variable "{variableName}".""")
        variableDataType = DATA_TYPES_DICT[variableName]

        # Get old set of IDs
        logging.info("""    Getting the old set of IDs.""")
        oldMap = oldMaps[variableName]
        oldMap = pd.DataFrame(oldMap)
        if len(oldMap) > 0:
            if variableDataType.lower() == "numeric":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.unique()
                oldIDSet = set([ditchFloat(el) for el in oldIDSet])  # NOTE: Hack. Convert values to type int or string
            elif variableDataType.lower() == "string":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.astype(str)
                oldIDSet = oldIDSet.unique()
            elif variableDataType.lower() == "numeric_or_string":
                oldIDSet = oldMap[variableName]
                oldIDSet = pd.Series(oldIDSet)
                oldIDSet = oldIDSet.astype(str)
                oldIDSet = oldIDSet.unique()
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)
        elif len(oldMap) == 0:
            oldIDSet = set()
        logging.info(f"""    The size of this set is {len(oldIDSet):,}.""")

        # Get new set of IDs
        newSetPath = getIDValuesOutput.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]}.txt")
        logging.info(f"""    Getting the new set of IDs from "{newSetPath.absolute().relative_to(rootDirectory)}".""")
        try:
            setSeries = pd.read_table(newSetPath, header=None)[0]
            newIDSet = set(setSeries.to_list())
        except EmptyDataError as err:
            _ = err
            newIDSet = set()
        if variableDataType.lower() == "numeric":
            newIDSet = set([ditchFloat(el) for el in newIDSet])  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            newIDSet = set([str(el) for el in newIDSet])
        elif variableDataType.lower() == "numeric_or_string":
            newIDSet = set([str(el) for el in newIDSet])
        else:
            msg = "The table column is expected to have a data type associated with it."
            logging.error(msg)
            raise Exception(msg)
        logging.info(f"""    The size of this set is     {len(newIDSet):,}.""")

        # Set difference
        IDSetDiff = newIDSet.difference(oldIDSet)
        logging.info(f"""    The set difference size is  {len(IDSetDiff):,}.""")
        valuesToMap[variableName] = IDSetDiff

        # Save new subset to `setsToMapDataDir`
        fpath = setsToMapDataDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]}.JSON")
        with open(fpath, "w") as file:
            if variableDataType.lower() == "numeric":
                li = [ditchFloat(IDNumber) for IDNumber in IDSetDiff]  # NOTE: Hack. Convert values to type int or string
            elif variableDataType.lower() == "string":
                li = list(IDSetDiff)
            elif variableDataType.lower() == "numeric_or_string":
                li = list(IDSetDiff)
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)
            file.write(json.dumps(li, default=handleDatetimeForJson))
        if len(IDSetDiff) == 0:
            series = pd.Series(dtype=int)
        else:
            if variableDataType.lower() == "numeric":
                series = pd.Series(sorted(list(IDSetDiff)))
            elif variableDataType.lower() == "string":
                series = pd.Series(sorted([str(el) for el in IDSetDiff]))
            elif variableDataType.lower() == "numeric_or_string":
                series = pd.Series(sorted([str(el) for el in IDSetDiff]))
            else:
                msg = "The table column is expected to have a data type associated with it."
                logging.error(msg)
                raise Exception(msg)

    # Get numbers for new map
    logging.info("""Getting numbers for new map.""")
    newNumbersDict = {}
    for variableName in collectedVariables:
        oldMap = oldMaps[variableName]
        if len(oldMap) > 0:
            oldNumbersSet = set(oldMap.iloc[:, 1].values)
        elif len(oldMap) == 0:
            oldNumbersSet = set()
        # Get quantity of numbers needed for map
        quantityOfNumbersUnmapped = len(valuesToMap[variableName])
        # Get new numbers
        lenOlderNumbersSet = len(oldNumbersSet)
        if lenOlderNumbersSet == 0:
            newNumbers = list(range(1, quantityOfNumbersUnmapped + 1))
        else:
            newNumbersSet = makeSetComplement(oldNumbersSet, quantityOfNumbersUnmapped)
            newNumbers = sorted(list(newNumbersSet))
        newNumbersDict[variableName] = newNumbers

    # Map un-mapped values
    logging.info("""Mapping un-mapped values.""")
    for file in setsToMapDataDir.iterdir():
        variableName = FILE_NAME_TO_VARIABLE_NAME_DICT[file.stem]
        variableDataType = DATA_TYPES_DICT[variableName]
        logging.info(f"""  Working on un-mapped values for variable "{variableName}" located at "{file.absolute().relative_to(rootDirectory)}".""")
        # Map contents
        values = valuesToMap[variableName]
        if variableDataType.lower() == "numeric":
            values = set(int(float(value)) for value in values)  # NOTE: Hack. Convert values to type int or string
        elif variableDataType.lower() == "string":
            pass
        elif variableDataType.lower() == "numeric_or_string":
            pass
        else:
            msg = "The table column is expected to have a data type associated with it."
            logging.error(msg)
            raise Exception(msg)
        numbers = sorted(list(newNumbersDict[variableName]))
        deIdIDSuffix = VARIABLE_SUFFIXES[variableName]["deIdIDSuffix"]
        map_ = makeMap(IDset=values,
                       IDName=variableName,
                       startFrom=numbers,
                       irbNumber=IRB_NUMBER,
                       suffix=deIdIDSuffix,
                       columnSuffix=variableName,
                       deIdentificationMapStyle="lemur",
                       logger=logging.getLogger())
        # Save map
        mapPath = runOutputDir.joinpath(f"{VARIABLE_NAME_TO_FILE_NAME_DICT[variableName]} map.csv")
        map_.to_csv(mapPath, index=False)
        logging.info(f"""    De-identification map saved to "{mapPath.absolute().relative_to(rootDirectory)}".""")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # <<< End script body <<<
    logger.info(f"""Finished running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")



"""
Functions to convert files to the format required by DeepDe-ID.
"""

import logging
from pathlib import Path
from typing_extensions import (Dict,
                               List,
                               Union)

import pandas as pd

from drapi.code.drapi.drapi import readDataFile

def prep_for_text_deidentification(filepath: Path,
                                   output_directory: Path,
                                   logger: logging.Logger,
                                   log_file_name: bool = True,
                                   rename_columns: Union[Dict[Union[int, str], Union[int,str]],
                                                         List[Union[int, str]],
                                                         None] = None,
                                   columns_to_keep: Union[None, List[Union[int, str]]] = None) -> None:
    """
    """
    # Verbose block
    if log_file_name:
        logger.info(f"""Working on file "{filepath}".""")

    # Read file
    readerObject = readDataFile(fname=filepath,
                                engine="pyarrow")

    # Pre-process
    df = pd.DataFrame(readerObject)

    # Pre-process: Columns to keep
    if columns_to_keep:
        # Pre-process: Columns to keep: Assertions
        len_columns_to_keep = len(columns_to_keep)
        if len_columns_to_keep == 2:
            pass
        else:
            message = f"""The number of columns to keep must be precisely 2, got instead "{len_columns_to_keep:,}"."""
            logger.critical(message)
            raise Exception(message)

        # Pre-process: Columns to keep: Choose columns
        if all([isinstance(el, int) for el in columns_to_keep]):
            df = df.iloc[:, columns_to_keep]
        elif all([isinstance(el, str) for el in columns_to_keep]):
            df = df.loc[:, columns_to_keep]
        else:
            message = "We expect the values in `columns_to_keep` to be either all integers or all strings."
            logger.critical(message)
            raise Exception(message)

    # Rename columns
    if rename_columns:
        # Assertions: The number of columns to rename should be the same as the number of columns, if list-like, else it should be a dictionary-like.
        if isinstance(rename_columns, list):
            len_rename_columns = len(rename_columns)
            message = f"""If you wish to rename the columns, then `rename_columns` should have the same number of columns as your table."""
            if len_rename_columns == len(df.columns):
                pass
            else:
                logger.critical(message)
                raise Exception(message)
        elif isinstance(rename_columns, dict):
            pass
        else:
            raise Exception(f"""We expect `rename_columns` to be a list or dictionary. We got "{type(rename_columns)}" instead.""")
        
        # Use `rename_columns`
        if isinstance(rename_columns, list):
            for old_column, new_column in zip(df.columns, rename_columns):
                df = df.rename(columns={old_column: new_column})
        elif isinstance(rename_columns, dict):
            df = df.rename(columns=rename_columns)
        else:
            raise Exception(f"""We expect `rename_columns` to be a list or dictionary. We got "{type(rename_columns)}" instead.""")

    # Save as TSV
    file_stem = filepath.stem
    export_path = output_directory.joinpath(f"{file_stem}.TSV")
    df.to_csv(path_or_buf=export_path,
              index=False,
              sep="\t")



"""
Converts files to the format required by DeepDe-ID.

NOTE TODO There are TODO tags in this file.
"""

import argparse
import json
import logging
import multiprocessing as pp
import os
import pprint
import shutil
from functools import partial
from pathlib import Path
# Third-party packages
pass
# First-party packages
from drapi.code.drapi.prep_for_text_deidentification import prep_for_text_deidentification
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getTimestamp,
                                    loggingChoiceParser,
                                    makeDirPath)
from drapi.code.drapi.cli_parsers import parse_string_to_boolean
from drapi import loggingChoices
from drapi import __version__ as drapiVersion


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments
    parser.add_argument("--list_of_file_paths",
                        type=Path,
                        required=True,
                        nargs="+",
                        help="The list of file paths to convert.")

    parser.add_argument("--rename_columns",
                        type=json.loads)
    parser.add_argument("--log_file_name",
                        default=True,  # TODO Remove default value when you implement the use of `eval` elsewhere in this file.
                        type=parse_string_to_boolean)
    parser.add_argument("--columns_to_keep",
                        help="The indices or labels of columns to keep. Max 2.",
                        nargs="+")

    # Arguments: Meta-parameters
    parser.add_argument("--TIMESTAMP",
                        type=str)
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=loggingChoiceParser,
                        choices=loggingChoices,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    argNamespace = parser.parse_args()

    # Parsed arguments
    list_of_file_paths = argNamespace.list_of_file_paths
    rename_columns = argNamespace.rename_columns
    log_file_name = argNamespace.log_file_name
    columns_to_keep = argNamespace.columns_to_keep

    # Parsed arguments: Meta-parameters
    TIMESTAMP = argNamespace.TIMESTAMP
    LOG_LEVEL = argNamespace.LOG_LEVEL
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    # >>> Custom argument parsing: Parsing 1 >>>
    pass
    # <<< Custom argument parsing: Parsing 1 <<<

    # >>> Custom argument parsing: Parsing 2 >>>
    pass
    # <<< Custom argument parsing: Parsing 2 <<<
    # <<< Custom argument parsing <<<

    # >>> Argument checks >>>
    # NOTE TODO Look into handling this natively with `argparse` by using `subcommands`. See "https://stackoverflow.com/questions/30457162/argparse-with-different-modes"
    # >>> Argument checks: Check 1 >>>
    pass
    # <<< Argument checks: Check 1 <<<
    # >>> Argument checks: Check 2 >>>
    pass
    # <<< Argument checks: Check 2 <<<
    # <<< Argument checks <<<

    # Variables: Path construction: General
    if TIMESTAMP:
        runTimestamp = TIMESTAMP
    else:
        runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    currentWorkingDir = Path(os.getcwd()).absolute()
    projectDir = currentWorkingDir
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        intermediateDataDir = dataDir.joinpath("intermediate")
        outputDataDir = dataDir.joinpath("output")
        if intermediateDataDir:
            runIntermediateDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
        if outputDataDir:
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    # Variables: Path construction: Project-specific
    pass

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{choosePathToLog(path=thisFilePath, rootPath=projectDir)}".""")
    logger.info(f"""DRAPI-Lemur version is "{drapiVersion}".""")
    logger.info(f"""All other paths will be reported in debugging relative to the current working directory: "{choosePathToLog(path=projectDir, rootPath=projectDir)}".""")

    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)  # TODO Remove secrets from list to print, e.g., passwords.
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin script body >>>

    # Parallel implementation
    with pp.Pool() as pool:
        # NOTE TODO For developer: should not pass implicitely-valued None-types to the function. I should handle this better, maybe by using `eval`. See https://www.geeksforgeeks.org/execute-string-code-python/
        results = pool.map(partial(prep_for_text_deidentification,
                                   output_directory=runOutputDir,
                                   logger=logger,
                                   log_file_name=log_file_name,
                                   rename_columns=rename_columns,
                                   columns_to_keep=columns_to_keep),
                           list_of_file_paths)

    # <<< End script body <<<

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")

    # Remove intermediate files, unless running in `DEBUG` mode.
    if logger.getEffectiveLevel() > 10:
        logger.info("Removing intermediate files.")
        shutil.rmtree(runIntermediateDir)
        logger.info("Removing intermediate files - done.")

    # Script end confirmation
    logger.info(f"""Finished running "{choosePathToLog(path=runOutputDir, rootPath=projectDir)}".""")



"""
Quality Assurance functions to make sure columns were properly converted or de-identified.
"""

import logging
# Third-party packages
import pandas as pd
# First-party packages
pass

# Functions

def spotCheckColumns(listOfFiles1: list,
                     listofFiles2: list,
                     connectionString: str,
                     logger: logging.Logger,
                     moduloChunks: int,
                     chunkSize: int = 50_000,
                     messageModuloChunks: int = 50,
                     messageModuloFiles: int = 100,
                     columnsToCheck1: dict[int: str] = {0: "visit_occurrence_id",
                                                        1: "person_id",
                                                        2: "provider_id"},
                     columnsToCheck2: dict[int: str] = {0: "Encounter # (CSN)",
                                                        1: "Patient Key",
                                                        2: "Provider Key"}) -> None:
    """
    This function "spot checks" pairs of files to see if the passed variables have been properly converted from the OMOP data model to the UF Health data model.

    In `columnsToCheck1` and `columnsToCheck2` the keys of the dictionary have the following meaning:
    - 0: Encounter ID Variable Name
    - 1: Patient ID Variable Name
    - 2: Provider ID Variable Name

    The code is only meant to check these three ID types, because under the hood these integer values are mapped to SQL queries. So, for example, you could not pass `{0: "Note Key"}` to the function, because it would use the SQL query that maps encounter IDs.

    NOTE This has only been tested and verified on 3-on-3 checks. That is, checking if all three informative OMOP ID variables have been mapped to their three UF Health equivalents. The code should be able to also do spot checks for less number of variables, like just `person_id` to `Patient Key`, but it has not been tested.
    """
    with open("../Concatenated Results/sql/BO Variable Spot Check - Encounter # (CSN).sql", "r") as file:
        query0_0 = file.read()
    with open("../Concatenated Results/sql/BO Variable Spot Check - Patient Key.sql", "r") as file:
        query1_0 = file.read()
    with open("../Concatenated Results/sql/BO Variable Spot Check - Provider Key.sql", "r") as file:
        query2_0 = file.read()

    for file1, file2 in zip(listOfFiles1, listofFiles2):
        try:
            logger.info(f"""  Working on file pair:
        `file1`: "{file1}"
        `file2`: "{file2}".""")
            df2_0 = pd.read_csv(file2, nrows=2)
            df2_0columns = df2_0.columns
            columnChecks2 = df2_0columns.isin(columnsToCheck2.values())
            logger.debug(f"""  Group 2 table:\n{df2_0.to_string()}""")
            logger.debug(f"""  Group 2 columns:\n{df2_0.columns}""")
            logger.debug(f"""  Group 2 columns check:\n{columnChecks2}""")
            spot2df = df2_0.loc[0, columnChecks2]
            spot2df = pd.DataFrame(spot2df)  # For type hinting
            logger.info(f"  A row has been selected for a spot check:\n{spot2df.T.to_string()}")
            # Convert group 2 spot to group 1 values
            spot2dict = spot2df.to_dict()[0]
            logger.info(spot2dict)
            query0_1 = query0_0[:]
            query1_1 = query1_0[:]
            query2_1 = query2_0[:]
            spot1dict = {}
            if 0 in columnsToCheck2.keys():
                encounterIDVariableName = columnsToCheck2[0]
                if df2_0columns.isin([encounterIDVariableName]).sum() > 0:
                    queryEncounterIDValue = spot2dict[encounterIDVariableName]
                    query0_1 = query0_1.replace("{PYTHON_VARIABLE: Encounter # (CSN)}", str(queryEncounterIDValue))
                    result0 = pd.read_sql(sql=query0_1, con=connectionString)
                    if result0.shape != (1,2):
                        logger.warning(f"""  ..  The encounter ID mapping is not one-to-one:\n{result0.to_string()}.""")
                        spot1EncounterID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1EncounterID = result0["visit_occurrence_id"][0]
                    spot1dict["visit_occurrence_id"] = [spot1EncounterID]
            if 1 in columnsToCheck2.keys():
                patientIDVariableName = columnsToCheck2[1]
                if df2_0columns.isin([patientIDVariableName]).sum() > 0:
                    queryPatientIDValue = spot2dict[patientIDVariableName]
                    query1_1 = query1_1.replace("{PYTHON_VARIABLE: Patient Key}", str(queryPatientIDValue))
                    result1 = pd.read_sql(sql=query1_1, con=connectionString)
                    if result1.shape != (1,2):
                        logger.warning(f"""  ..  The patient ID mapping is not one-to-one:\n{result1.to_string()}.""")
                        spot1PatientID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1PatientID = result1["person_id"][0]
                    spot1dict["person_id"] = [spot1PatientID]
            if 1 in columnsToCheck2.keys():
                providerIDVariableName = columnsToCheck2[2]
                if df2_0columns.isin([providerIDVariableName]).sum() > 0:
                    queryProviderIDValue = spot2dict[providerIDVariableName]
                    query2_1 = query2_1.replace("{PYTHON_VARIABLE: Provider Key}", str(queryProviderIDValue))
                    result2 = pd.read_sql(sql=query2_1, con=connectionString)
                    if result2.shape != (1,2):
                        logger.warning(f"""  ..  The provider ID mapping is not one-to-one:\n{result2.to_string()}.""")
                        spot1ProviderID = None  # NOTE This is a dummy value assigned to the variable to allow the code to continue running, but at the same time to indicate that an unexpected or undesired result has occurred.
                    else:
                        spot1ProviderID = result2["provider_id"][0]
                    spot1dict["provider_id"] = [spot1ProviderID]
            spot1Targetdf = pd.DataFrame.from_dict(data=spot1dict, orient="columns")
            logger.info(f"""The values that we are looking for in the group 1 set of files is below:\n{spot1Targetdf}""" )
            # Chunk
            chunkGenerator1 = pd.read_csv(file1, chunksize=chunkSize, low_memory=False)
            chunkGenerator2 = pd.read_csv(file1, chunksize=chunkSize, low_memory=False)
            logger.info("    Counting the number of chunks in the file.")
            it1Total = sum([1 for _ in chunkGenerator1])
            logger.info(f"    Counting the number of chunks in the file - done. There are {it1Total:,} chunks.")
            if it1Total < messageModuloChunks:
                moduloChunks = it1Total
            else:
                moduloChunks = round(it1Total / messageModuloChunks)
            if it1Total / moduloChunks < 100:
                moduloChunks = 1
            else:
                pass
            for it1, df1Chunk in enumerate(chunkGenerator2, start=1):
                if it1 % moduloChunks == 0:
                    allowLogging = True
                else:
                    allowLogging = False
                if allowLogging:
                    logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total:,}.""")
                df1Chunk = pd.DataFrame(df1Chunk)  # For type hinting
                columnChecks1 = df1Chunk.columns.isin(columnsToCheck1.values())
                df1 = df1Chunk.loc[:, columnChecks1]
                df1 = pd.DataFrame(df1)  # For type hinting
                mask = df1.isin(spot1Targetdf.values.flatten()).all(axis=1)
                rowLocation = mask.index[mask][0]
                if mask.sum() > 0:
                    spot1Hitdf = df1[mask]
                    logger.info(f"""    Found spot located at index value (row number) "{rowLocation}":\n{spot1Hitdf.to_string()}.""")
                    break
        except Exception as error:
            logger.fatal(error)